;; Functions for representing and selecting subcubes in the CM2
;; File:  cubes.lisp

(in-package '*lisp)

;; NOTE:  The seemingly superfluous ash!! [instead of (!! (ash ...))]
;;        permits large fields.

(*proclaim '(type (field-pvar 24) self))

(*defvar self  (self-address!!))   ; for convenience

;; Description of current geometry (as cube):

(*proclaim '(type fixnum addrsz physz routsz chipsz virtsz))

(defvar addrsz 15)     ;; Number of bits in virtual cube address
(defvar physz  15)     ;; Number of bits in physical cube address
(defvar routsz 11)     ;; Number of off-chip (router) address bits
(defvar chipsz  4)     ;; Number of on-chip address bits (not virt.)
(defvar virtsz  0)     ;; Number of virtual address bits

;; These satisfy    addrsz = physz  + virtsz
;;  and             physz  = routsz + chipsz
;; chipsz = 4 in all current CM models

;; Utility functions for reading and displaying addresses, face descriptions

;; Read line as bit pattern.  Leading bit is used to fill high-order
;; positions not specified, extra high-order bits are ignored.
;; Type no spaces.  Returns pattern as integer < 2^addrsz.

(defun read-bits ()
 (format t " (bits)> ")
 (let* ((ptrn (read-line))    (pn (length ptrn))
        (lc (schar ptrn 0))   (lb (if (char= lc #\0) 0 1))
        (bt (ash 1 addrsz))   (x 0)
        (n addrsz))
   (if (< pn n)
     (let ((w bt))
       (setq bt (ash 1 pn))
       (if (= lb 1) (setq x (- w bt)))
       (setq n pn)))
   (dotimes (i n)
     (setq bt (ash bt -1))
     (if (char= (schar ptrn (+ i (- pn n))) #\1) (setq x (logior x bt))))
   x
 ) ; end let*
)

;; Read line as face description.  Leading symbol is used to fill high-order
;; positions not specified, extra high-order symbols are ignored.
;; Type no spaces.  Returns [beginning, pattern] as integers.

(defun read-desc ()
 (format t " (dscr)> ")
 (let* ((ptrn (read-line))    (pn (length ptrn))
        (lc (schar ptrn 0))   (bt (ash 1 addrsz))
        (bg 0)     (pt 0)     (n addrsz))
   (if (and (< pn n) (char/= lc #\*))
     (let ((w bt))
       (setq bt (ash 1 pn))
       (setq pt (- w bt))
       (if (char= lc #\1) (setq bg pt))
       (setq n pn)))
   (dotimes (i n)
     (setq bt (ash bt -1))
     (setq lc (schar ptrn (+ i (- pn n))))
     (when (char/= lc #\*)
        (setq pt (logior pt bt))
        (if (char= lc #\1) (setq bg (logior bg bt)))))
   (values bg pt)
 ) ; end let*
)

;; Compute number of beginning and ending spaces for k-character heading
;; over c columns (2 characters per column).

(defun p-center (k c)
 (let* ((ln (- (ash c 1) k 1)) (e (ash ln -1)) (b (- ln e)))
   (values b e))
)

;; Print column separator to follow column i, either space or "|".
;; Returns i (its argument).

(defun p-sepr (i)
 (write-char
  (if (or (= i routsz) (= i addrsz) (= i (+ routsz chipsz))) #\| #\space))
 i
)

;; Print heading for addresses, descriptions based on current configuration
;; information.  Output stream should be at beginning of line.

(defun p-head()
 (format t "  ")
 (multiple-value-bind (b e) (p-center 8 routsz)
   (format t "|~v@Toff-chip~v@T" b e))
 (multiple-value-bind (b e) (p-center 7 chipsz)
   (format t "|~v@Ton-chip~v@T" b e))
 (if (plusp virtsz)
   (multiple-value-bind (b e) (p-center 1 virtsz)
     (format t "|~v@Tv~v@T" b e)))
 (format t "|~%  |")
 (do ((i 0 (p-sepr (1+ i))) (n addrsz))
     ((= i addrsz))
     (decf n)
     (format t "~1D" (truncate n 10)))
 (format t "~%  |")
 (do ((i 0  (p-sepr (1+ i))) (n addrsz))
     ((= i addrsz))
     (decf n)
     (format t "~1D" (mod n 10)))
 (format t "~%  |")
 (do ((i 0  (p-sepr (1+ i))))
     ((= i addrsz))
     (write-char #\-))
 (format t "~%")
)

;; Print bit pattern.  Output stream should be at beginning of line.
;; No following newline.

(defun p-bits (b)
 (format t "  |")
 (do ((i 0 (p-sepr (1+ i))) (bt (ash 1 addrsz)))
     ((= i addrsz))
     (setq bt (ash bt -1))
     (write-char (if (zerop (logand bt b)) #\0 #\1))
 )
)

;; Print description given by be, pt.  Output stream should be at
;; beginning of line.  No following newline.

(defun p-desc (be pt)
 (format t "  |")
 (do ((i 0 (p-sepr (1+ i))) (bt (ash 1 addrsz)))
     ((= i addrsz))
     (setq bt (ash bt -1))
     (write-char
        (if (zerop (logand bt pt)) #\*
         (if (zerop (logand bt be)) #\0 #\1)))
 )
)

;; Setup for geometry descriptions, etc.

(defun setup-cubes ()
 (setq addrsz (logcount (1- *number-of-processors-limit*)))
 (setq physz  (logcount (1- cm:*physical-processors-limit*)))
 (setq virtsz (- addrsz physz))
 (setq routsz (- physz chipsz))
 (format t "~%2^~D vprocs on 2^~D procs, v/p = ~D.~%"
               addrsz         physz    (ash 1 virtsz))
)
